/*
 * $Id: SVGGraphics.java,v 1.1 2000/04/26 14:52:59 ivan Exp $
 */
/*--------------------------------------------------------------------------------
 Copyright Stichting Mathematisch Centrum 2000

  Permission to use and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice, author statement and this permission,
  warranties and liability notice appear in all copies of this software
  and related documentation.

  Warranties
  Stichting Mathematisch Centrum makes no warranties, express or implied, with
  regard to this software. In particular, but without limitation, SMC makes
  no warranty of fitness for a particular use, no warranty of freedom from
  infringement of any patent or copyright, nor any warranty as to accuracy.
  Accordingly, SMC assumes no obligation to furnish assistance of any kind
  whatsoever or to furnish additional information or documentation.

  Liability
  In no event shall Stichting Mathematisch Centrum be liable for any damages
  arising out of or in connection with the use or performance of this
  software.
*/
package cwi.SVGGraphics;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.awt.image.renderable.*;
import java.awt.font.*;
import java.io.*;
import java.lang.*;
import java.util.*;
import java.text.*;
/**
* SVG generator for Java.
* <p>
* Using this class makes it easy(er) to generate SVG files from Java. The intention is to
* use SVG as a kind of "screen dump", using possibly the same user methods as for the 
* generation of the screen content. This is the reason why the object is defined as 
* an extension of <code>Graphics2D</code>: various methods which can use any
* <code>Graphics2D</code> instances can be used in conjunction with this class. The 
* approach is similar to the printing facilities in Java.
* Here is a typical usage of the class:
* <pre>
*      SVGGraphics svg = SVGGraphics.create(yourCanvas);
*      if( svg == null ) return;
*      yourCanvas.yourDrawMethod(svg);
*      svg.close();
* </pre>
* Note the usage of the <code>close</code> method, which is necessary to properly close
* the SVG file.
* <p>
* Most of the graphics calls in <code>Graphics2D</code> translate into SVG easily.
* There are some differences, difficulties, or not-yet-implemented features, though, check the
* comments on the individual methods themselves.
* The generated SVG file includes a number of groups: one for each
* different transformation and/or stroke settings. Gradient colours of Java are 
* separated into a distinct set of definitions.
* <p>
* Coordinates in the generated SVG file are simply passed on from Java,
* and the transformation matrices in the groups take care of the proper positioning. The only
* exception are the coordinates in the linear gradients in the definitions, where the 
* coordinates are transformed before being stored in the file. 
* <p>
* As mentioned before most of the Java calls can be translated into SVG easily. The only
* major exception is the handling of images. In Java, images are pixels (by the time the 
* methods in <code>Graphics</code> are called) whereas SVG refers to URL-s. I did not see any
* obvious ways to store pixel information in SVG and, even if it was possible, it would be
* a waste of bandwidth. The solution chosen in <code>SVGGraphics</code> is to accept
* only those <code>Image</code> objects, whose property with a key <code>"url"</code> has been set.
* This url is then used as a reference in the SVG file. 
* The only problem with using <code>"url"</code> key for an image that it is not immediately
* obvious how one would set a property on an image. For reasons which I do not completely understand,
* although there is a public <code>getProperty</code> method for an <code>Image</code> object,
* there is no public <code>setProperty</code> method. The only way I found was to use a special
* image consumer filter to generate a new <code>Image</code> instance with the property set. To
* make life easier for the users of <code>SVGGraphics</code>, a separate class, called
* {@link SVGImageFilter <code>SVGImageFilter</code>}, is available. See the description of that class for further details
* on how to use it. Note that the same restriction applies for the texture paint: the buffered image 
* which is used in a texture paint has to include the <code>"url"</code> property. If it is there,
* a separate pattern specification (in SVG terms) is generated into the output.
* <p>
* The user might also want to look at the class 
* {@link STest STest} which 
* I used for testing <code>SVGGraphics</code>.
*
* @see SVGImageFilter
* @author Ivan Herman 
* @version 1.0
*/
public class SVGGraphics extends Graphics2D {
  private boolean debug = true;
  
  /** 
  * The target of the SVG creation. By default, this is set to System.out, but
  * can (and should) be set by the creation process.
  */
  private PrintStream out = System.out;
  /**
  * Width and height of the canvas, used to set the dimensions in the svg element. It is usually
  * set using the dimensions of the 'parent' component, or is set by the user.
  */
  private int         width, height;
  /**
  * "Parent component". Is used to extract device characteristics, for example. It does not have
  * any other 'active' role in the svg generation.
  */
  private Component   theComponent = null;
  
  //===============================================================================
  // Operating variables
  //===============================================================================
    
  /**
  * String buffer collecting the svg statements in a <defs> element. 
  */
  private StringBuffer defs = new StringBuffer("");
  /**
  * String buffer collecting the svg statements in the core of the file, ie, outside of the 
  * <defs> elements.
  */
  private StringBuffer core = new StringBuffer();
  
  /*
  * Space level for the core part. Number of spaces to be printed in a line for a correct indendation
  */
  private int     sp           = 0;
  /*
  * Space level for the definition part. Number of spaces to be printed in a line for a correct indendation
  */
  private int     spd          = 0;
  /**
  * Whether we are inside a <g> or not. Most of the time we are, except for the start.
  */
  private boolean inGroup      = false;
  /**
  * Current length of the core buffer. Is used by the optimization on getting empty <g> elements out
  * by deleting it from the core buffer the unnecessary parts.
  */
  private int     bufferLength = 0;
  
  /**
  * Print a sequence of lines to the output. Users can use this method to put svg statements directly into the
  * output stream. Note, however, that no semantic check is done on the string, so user have
  * to know what they are doing!
  * @param arg texts to be put appende to the output stream
  */
  public void println(String[] arg)
  {
    for( int i = 0; i < arg.length; i++ )
      println(arg[i]);
  }
  
  /**
  * Print a line to the output. Users can use this method to put SVG statements directly into the
  * output stream. Note, however, that no semantic check is done on the string, so user have
  * to know what they are doing!
  * @param arg text to be put appende to the output stream
  */
  public void println(String arg)
  {
    for( int i = 0; i < sp; i++ )
      core.append(" ");
    core.append(arg);
    core.append("\n");
    bufferLength = core.length();
  }
  
  /**
  * Print a line to the defintion section. 
  * @param arg text to be put append to the output stream
  */
  public void printlnDef(String arg)
  {
    for( int i = 0; i < spd; i++ )
      defs.append(" ");
    defs.append(arg);
    defs.append("\n");
  }

  //===============================================================================
  // SVG state variables
  //===============================================================================
    
  // Current transformation. The transformation is stored for comparisons and for the
  // transformation of tags in a <defs> section. Otherwise, we rely on the transformations
  // in SVG.
  private AffineTransform trans = new AffineTransform();

  // possible paint types currently defined in Java2D
  private static final int PAINT_COLOR    = 0;
  private static final int PAINT_GRADIENT = 1;
  private static final int PAINT_TEXTURE  = 2;

  private int current_paint_type    = PAINT_COLOR;

  // Current paint values. Note that I have not implemented the equivalent of the paint
  // texture!  
  private Color         backgroundColour  = Color.white;
  private Color         currentColour     = null;
  private GradientPaint currentGradient   = null;
  private TexturePaint  currentTexture    = null;

  private BasicStroke    currentStroke = new BasicStroke();
  private AlphaComposite currentComposite = null;
  private Font           currentFont = null;

  //===============================================================================
  // SVG is a set of strings, so there are a load of internal utilities to produce
  // strings in svg formats to to be put into the target

  // Cut the float value to two significant digits
  private String flstr(float f)
  {
    return Float.toString(((float)Math.round(f*100.0f))/100.0f);
  }
    
  // Cut the float value to two significant digits
  private String flstr(double f)
  {
    return Float.toString(((float)Math.round(f*100.0))/100.0f);
  }
  
  /**
  * Beware! This version is used for linear gradients, and it does an on-the-fly
  * transformation of the data!
  */
  private String pairToString(Point2D p1, Point2D p2)
  {
    Point2D tp1 = trans.transform(p1,null);
    Point2D tp2 = trans.transform(p2,null);
    StringBuffer retval = new StringBuffer();
    retval.append("x1=\"" + flstr(tp1.getX()) + "\" y1=\"" + flstr(tp1.getY()) + "\" ");
    retval.append("x2=\"" + flstr(tp2.getX()) + "\" y2=\"" + flstr(tp2.getY()) + "\" ");
    return retval.toString();
  }

  private String pairToString(float x1, float y1, float x2, float y2)
  {
    StringBuffer retval = new StringBuffer();
    retval.append("x1=\"" + flstr(x1) + "\" y1=\"" + flstr(y1) + "\" ");
    retval.append("x2=\"" + flstr(x2) + "\" y2=\"" + flstr(y2) + "\" ");
    return retval.toString();
  }
  
  private String pointToString(float x, float y)
  {
    return "x=\"" + flstr(x) + "\" y=\"" + flstr(y) + "\" ";
  }

  private String ovalToString(float x, float y, float width, float height)
  {
    StringBuffer retval = new StringBuffer();
    retval.append("cx=\"" + flstr(x) + "\" ");
    retval.append("cy=\"" + flstr(y) + "\" ");
    retval.append("rx=\"" + flstr(width) + "\" ");
    retval.append("ry=\"" + flstr(height) + "\" ");
    return retval.toString();
  }

  private String rectToString(float x, float y, float width, float height)
  {
    StringBuffer retval = new StringBuffer();
    retval.append("x=\"" + flstr(x) + "\" ");
    retval.append("y=\"" + flstr(y) + "\" ");
    retval.append("width=\"" + flstr(width) + "\" ");
    retval.append("height=\"" + flstr(height) + "\" ");
    return retval.toString();
  }

  private String rectToStringTrans(float x, float y, float width, float height)
  {
    Point2D p1 = new Point2D.Float(x,y);
    Point2D p2 = new Point2D.Float(x+width,y+height);
    Point2D tp1 = trans.transform(p1,null);  
    Point2D tp2 = trans.transform(p2,null);  
    StringBuffer retval = new StringBuffer();
    retval.append("x=\"" + flstr(tp1.getX()) + "\" ");
    retval.append("y=\"" + flstr(tp1.getY()) + "\" ");
    retval.append("width=\"" + flstr(tp2.getX()-tp1.getX()) + "\" ");
    retval.append("height=\"" + flstr(tp2.getY()-tp1.getY()) + "\" ");
    return retval.toString();
  }

  private String arcToString(float x, float y, float width, float height,
                             float startAngle, float arcAngle,
                             boolean close)
  {
    double alpha = Math.toRadians((double)startAngle);
    double beta  = Math.toRadians((double)arcAngle);
    
    double x1 = x + width *Math.cos(alpha);
    double y1 = y + height*Math.sin(alpha);
    double x2 = x + width*Math.cos(alpha+beta);
    double y2 = y + height*Math.sin(alpha+beta);
    
    StringBuffer buff = new StringBuffer();
    buff.append("d=\"");
    buff.append((close? "L" : "M") + flstr(x1) + " " + flstr(y1) + " ");
    buff.append("A" + flstr(width)  + " " + flstr(height)  + " 0");
    buff.append( ((Math.abs(arcAngle) >= 180.0) ? " 1 " : " 0 ") );
    buff.append( ((arcAngle > 0.0) ? " 1 " : " 0 ") );
    buff.append( flstr(x2) + " " + flstr(y2));
    if( close ) buff.append(" Z");
    buff.append("\"");
    return buff.toString();
  }

  private String pointsToString(int xPoints[], int yPoints[], int nPoints)
  {
    StringBuffer buff = new StringBuffer();
    buff.append("points=\"");
    for( int i = 0; i < nPoints; i++ ) {
      buff.append(flstr((float)xPoints[i]) + "," + flstr((float)yPoints[i]) + " ");
    }
    buff.append("\"");
    return buff.toString();
  }
  
  private String shapeToString(Shape s, boolean close)
  {
    StringBuffer retval = new StringBuffer("d=\"");
    PathIterator points = s.getPathIterator(null);
    boolean closed = false;
    
    for( ; points.isDone() == false ; points.next() ) {
      float p[] = new float[6];
      int mode = points.currentSegment(p);
      switch( mode ) {
        case PathIterator.SEG_MOVETO : retval.append("M" + flstr(p[0]) + " " + flstr(p[1]) + " ");
                                       break;
        case PathIterator.SEG_LINETO : retval.append("L" + flstr(p[0]) + " " + flstr(p[1]) + " ");
                                       break;
        case PathIterator.SEG_QUADTO : retval.append("Q" + flstr(p[0]) + " " + flstr(p[1]));
                                       retval.append(" "  + flstr(p[2]) + " " + flstr(p[3]));
                                       break;
        case PathIterator.SEG_CUBICTO: retval.append("C" + flstr(p[0]) + " " + flstr(p[1]));
                                       retval.append(" "  + flstr(p[2]) + " " + flstr(p[3]));
                                       retval.append(" "  + flstr(p[4]) + " " + flstr(p[5]));
                                       break;
        case PathIterator.SEG_CLOSE  : retval.append("Z");
                                       closed = true;
                                       break;
      }
    }
    
    retval.append(( (close && !closed)? " Z\"" : "\""));
    return retval.toString();
  }

  // This method tries to optimize in the sense that the defaults of SVG are not
  // explicitly set. This method is used for the generation of a new group, ie, it
  // always create a new 'state', so this can be done without any danger.
  private String strokeToString()
  {
    if( currentStroke == null ) {
      return "";
    } else {
      StringBuffer buff = new StringBuffer();

      if( currentStroke.getLineWidth() != 1.0f )
        buff.append("stroke-width:" + flstr(currentStroke.getLineWidth()) + "; ");
      
      // Check for the default case...  
      if( currentStroke.getEndCap() != BasicStroke.CAP_BUTT ) { 
        // This is not the default...   
        buff.append("stroke-linecap:");
        switch( currentStroke.getEndCap() ) {
          case BasicStroke.CAP_ROUND:  buff.append("round; ");  break;
          case BasicStroke.CAP_SQUARE: buff.append("square; "); break;
        }
      }
      
      // Check for the default case...  
      if( currentStroke.getEndCap() != BasicStroke.JOIN_MITER ) {
        buff.append("stroke-linejoin:");
        switch( currentStroke.getEndCap() ) {
          case BasicStroke.JOIN_BEVEL: buff.append("bevel; "); break;
          case BasicStroke.JOIN_ROUND: buff.append("round; "); break;
        }
      } else {
        if( currentStroke.getMiterLimit() != 8.0f )
          buff.append("stroke-mitterlimit:" + currentStroke.getMiterLimit() + "; ");
      }

      float[] dash = currentStroke.getDashArray();
      if( dash != null && dash.length != 0 ) {
        buff.append("stroke-dasharray:");
        for( int i = 0; i < dash.length; i++ ) {
          buff.append(" " + dash[i]);
        }
        buff.append("; ");
        if( currentStroke.getDashPhase() != 0.0f ) {
          buff.append(" stroke-dashoffset:" + currentStroke.getDashPhase() + "; ");
        }
      }

      return buff.toString();
    }
  }

  // Why is it that the the 'basic' (i.e., identifiable by name) colours of Java are different than
  // that of SVG? One might think that these things would be standardised by now...
  private String colToString(Color c)
  {
    if( c == null ) {
      return "";
    } else if( c.equals(Color.black) == true ){
      return "black";
    } else if( c.equals(Color.blue) == true ) {
      return "blue";
    } else if( c.equals(Color.gray) == true ) {
      return "gray";
    } else if( c.equals(Color.green) == true ) {
      return "green";
    } else if( c.equals(Color.red) == true ) {
      return "red";
    } else if( c.equals(Color.white) == true ) {
      return "white";
    } else if( c.equals(Color.yellow) == true ) {
      return "yellow";
    } else {
      return "rgb("+c.getRed()+","+c.getGreen()+","+c.getBlue()+")";
    }
  }

  private String fillStyle()
  {
    StringBuffer buff = new StringBuffer();
    buff.append("style=\"stroke:none; ");
    String s = paintToString();
    if( !s.equals("") ) {
      buff.append("fill:" + s);
    }
    
    if( currentComposite != null ) {
      float alpha = currentComposite.getAlpha();
      if( alpha < 1.0 ) {
        buff.append("; fill-opacity:" + flstr(alpha));
      }
    }
      
    
    buff.append("\"");
    return buff.toString();
  }

  private String strokeStyle()
  {
    StringBuffer buff = new StringBuffer();
    buff.append("style=\"fill:none; ");
    String s = paintToString();
    if( !s.equals("") ) {
      buff.append("stroke:" + s + "; ");
    }
    //buff.append(strokeToString());
    if( currentComposite != null ) {
      float alpha = currentComposite.getAlpha();
      if( alpha < 1.0 ) {
        buff.append("; stroke-opacity:" + flstr(alpha));
      }
    }
    buff.append("\"");
    return buff.toString();
  }

  //===============================================================================
  // Group generation and control
  //===============================================================================
  private AffineTransform displayedTransform = new AffineTransform();
  private BasicStroke     displayedStroke    = new BasicStroke();
  private int groupBegin = 0;
  private int groupEnd   = 0;
  
  //
  // Generation of a new group. This is invoked after a change in the transformation
  // or the stroke.
  // The method implements two optimizations:
  // 1. if the transformation and the stroke did not change, the group is not generated
  //    In order to do this, it stores the previous transformation and stroke
  //    in the displayedTransform and displayedStroke variables, respectively.
  // 2. checks if the previous group was empty. This is done by storing the begin and end
  //    indeces of the core buffer for every group. If the stored end index and the
  //    current buffer index is identical, nothing has been added to the group. The 'delete'
  //    method of the StringBuffer is used on the 'core' buffer to delete the current group.
  private void newGroup() 
  {
    // see if a group change is necessary at all!
    if( trans.equals(displayedTransform)      == true &&
        currentStroke.equals(displayedStroke) == true ) 
      // An unnecessary change...
      return;
      
    displayedTransform = (AffineTransform)(trans.clone());
    displayedStroke    = new BasicStroke(currentStroke.getLineWidth(),currentStroke.getEndCap(),
                                         currentStroke.getLineJoin(),currentStroke.getMiterLimit(),
                                         currentStroke.getDashArray(),currentStroke.getDashPhase());
    
    if( inGroup ) {
      sp  -= 2;
      if( bufferLength == groupEnd ) {
        // There was an empty group!!!!
        core.delete(groupBegin,groupEnd);
      } else {
        // That was a real group
        println("</g>");
      }
    }
    // Store the group start index for the next round.
    groupBegin = core.length();
    
    StringBuffer str = new StringBuffer();
    str.append("<g transform=\"matrix(");
    double[] matrix = new double[6];
    trans.getMatrix(matrix);
    for( int i = 0; i < 6; i++ ) {
      str.append(flstr(matrix[i]));
      if( i < 5 ) str.append(",");
    }
    str.append(")\"");
    String stroke = strokeToString();
    if( stroke.equals("") == false ) {
      println(str.toString());
      str = new StringBuffer();
      str.append("    style=\"" + stroke + "\"");
    }
    str.append(">");
    println(str.toString());
    
    // Store the group end index for the next round.
    groupEnd = core.length();
    
    // increase indentation level
    sp += 2;  
    inGroup = true;
  }
     
  //============================================================================
  /**
  * Create a new <code>SVGGraphics</code> object.
  *
  * @param o output stream for the svg content
  * @param origin the "parent" component. Its bounds will be used to determine the
  * svg width and height parameters. If null, the values 400 and 400 will be used.
  * @return the new <code>SVGGraphics</code> object.
  */
  public static SVGGraphics create(PrintStream o, Component origin)
  {
    return new SVGGraphics(o,origin);
  }    
    
  /**
  * Create a new <code>SVGGraphics</code> object through a file picker widget. The user can select a file
  * through the widget, which will be used for output.
  *
  * @param origin the "parent" component. Its bounds will be used to determine the
  * svg width and height parameters. If null, the values 400 and 400 will be used.
  * @return the new SVGGraphics object. If the file choice was unsuccessful (e.g., the user
  * cancelled the choice), the returned value is null.
  */
  public static SVGGraphics create(Component origin)
  {
    PrintStream str = FilePicker.getSVGFile();
    if( str != null ) {
      return create(str,origin);
    } else {
      return null;
    }
  }
  
  //============================================================================
  
  private static String[] startUp = new String[] {
    "<?xml version=\"1.0\" encoding=\"iso-8859-1\" standalone=\"no\"?>",
    "<!-- Generated by cwi.SVGGraphics.SVGGraphics java object, copyright CWI, Amsterdam -->",
    "<!DOCTYPE svg SYSTEM \"svg-19991203.dtd\">"  
  };
  
  /**
  * Create a new object. In fact, this version is not used, it does not initialize 
  * anything; it is there for compatibility only with the superclasses.
  */
  protected SVGGraphics() {
      super();
  }
  
  /**
  * Create a new object. This is the constructor which is really used by this class: it
  * receives the output stream, and the size of the
  * picture, which should be put into the header. The SVG statements are printed through
  * the print stream.
  * Because the constructor is protected, it cannot be called directly from outside.
  * Users should call the {@link #create create} calls instead.
  * 
  * @param o output stream for the svg content
  * @param origin the "parent" component. Its bounds will be used to determine the
  * svg width and height parameters. If null, the values 400 and 400 will be used.
  * @see #create
  */
  protected SVGGraphics(PrintStream o, Component origin)
  {
    super();
    out = o;
    if( origin == null ) {
      width  = 400;
      height = 400;
    } else {
      theComponent = origin;
      Rectangle bounds = origin.getBounds();
      width  = bounds.width;
      height = bounds.height;
    }
    for( int i = 0; i < startUp.length; i++ )
      printlnDef(startUp[i]);
    printlnDef("<svg xml:space=\"preserve\" width=\"" + width + "\" height=\"" + height + "\">");
    sp  = 2;
    spd = 2;
    printlnDef("<defs>");
    spd = 4;
  }
  
  /**
  * Close the object. This will flush the buffers for the SVG file properly. It is 
  * imperative for the user to issue this call before closing the object, otherwise the
  * SVG file will not be generated!
  * @param closeFile if true, a 'close' command will also be issued on the <code>PrintStream</code>, too.
  */
  public void close(boolean closeFile)
  {
    if( inGroup ) {
      sp  -= 2;
      println("</g>");
    }
    sp  = 0;
    spd = 2;
    printlnDef("</defs>");
    out.print(defs.toString());
    out.print(core.toString());
    out.println("</svg>");
    out.flush();
    if( closeFile )
      out.close();
  }

  /**
  * Close the object. This will close the SVG file properly and flush the buffers. It is 
  * imperative for the user to issue this call before closing the object, otherwise the
  * SVG file will not be generated! The <code>PrintStream</code> object will also be closed.
  */
  public void close()
  {
    close(true);
  }

  /**
   * This method is inherited from the superclass. On its own 
   * the method issues simple {@link #close <code>close</code>}. In this sense, using this method is an alternative
   * to the <code>close</code> method.
   * @see #close 
   */
  public void dispose()
  {
    out.close();
  }

  /**
   * Creates a new <code>Graphics</code> object that is 
   * a copy of this <code>Graphics</code> object. This method is mandated by the
   * superclass but,
   * for the case of SVG, it does not
   * really make sense, because the internal buffers are not be copied. The call is
   * here to make the java compiler happy...
   * @return     a new graphics context that is a copy of this graphics context.
   */
  public Graphics create()
  {
    SVGGraphics g = new SVGGraphics(out,theComponent);
    return g;
  }

  //===============================================================================
  
  /**
   * Concatenates the current
   * <code>Graphics2D</code> <code>Transform</code> 
   * with a translation transform. A new group will be created in SVG.
   * @param tx the distance to translate along the x-axis
   * @param ty the distance to translate along the y-axis
   */
  public void translate(double tx, double ty) { trans.translate(tx,ty); }

  /**
   * Concatenates the current
   * <code>Graphics2D</code> <code>Transform</code>.  A new group will be created in SVG. 
   * @param  x   the <i>x</i> coordinate.
   * @param  y   the <i>y</i> coordinate.
   */
  public void translate(int x, int y) { translate((double)x, (double)y); newGroup(); }

  /**
   * Concatenates the current <code>Graphics2D</code>
   * <code>Transform</code> with a rotation transform.  A new group will be created in SVG.
   * x axis toward the positive y axis.
   * @param theta the angle of rotation in radians
   */
  public void rotate(double theta) { trans.rotate(theta); newGroup(); }

  /**
   * Concatenates the current <code>Graphics2D</code> 
   * <code>Transform</code> with a translated rotation 
   * transform.  A new group will be created in SVG.
   * @param theta the angle of rotation in radians
   * @param x,&nbsp;y coordinates of the origin of the rotation
   */
  public void rotate(double theta, double x, double y) { trans.rotate(theta,x,y); newGroup(); }

  /**
   * Concatenates the current <code>Graphics2D</code>
   * <code>Transform</code> with a scaling transformation.  A new group will be created in SVG. 
   * @param sx the amount by which X coordinates in subsequent
   * rendering operations are multiplied relative to previous
   * rendering operations.
   * @param sy the amount by which Y coordinates in subsequent 
   * rendering operations are multiplied relative to previous 
   * rendering operations.
   */
  public void scale(double sx, double sy) { trans.scale(sx,sy); newGroup(); }

  /**
   * Concatenates the current <code>Graphics2D</code>
   * <code>Transform</code> with a shearing transform.  A new group will be created in SVG.
   * @param shx the multiplier by which coordinates are shifted in 
   * the positive X axis direction as a function of their Y coordinate
   * @param shy the multiplier by which coordinates are shifted in
   * the positive Y axis direction as a function of their X coordinate
   */
  public void shear(double shx, double shy) { trans.shear(shx,shy); newGroup(); }

  /**
   * Composes an <code>AffineTransform</code> object with the 
   * <code>Transform</code> in this <code>Graphics2D</code> according 
   * to the rule last-specified-first-applied.  A new group will be created in SVG. 
   * @param Tx the <code>AffineTransform</code> object to be composed with 
   * the current <code>Transform</code>
   * @see #setTransform
   * @see AffineTransform
   */
  public void transform(AffineTransform Tx) { trans.concatenate(Tx); newGroup(); }
    
  /**
   * Sets the <code>Transform</code> in the <code>Graphics2D</code>
   * context.  A new group will be created in SVG.
   * @param Tx the <code>AffineTransform</code> object to be used in the
   * rendering process
   */
  public void setTransform(AffineTransform Tx) { trans = Tx; newGroup(); }

  /**
   * Returns a <em>copy</em> of the current <code>Transform</code> in the 
   * <code>Graphics2D</code> context.
   * @return the current <code>AffineTransform</code> in the 
   *             <code>Graphics2D</code> context.
   */
  public AffineTransform getTransform() { return (AffineTransform)trans.clone(); }


  //==================================================================================
  // These parameters are used to identify the "current" linear gradient.
  // A possible optimization in the future is to try to identify whether
  // two elements "share" the same linear gradient; the current version 
  // blindly generates linear gradients for each call of <code>setPaint</code>.
  private static String grPrefix          = "Gradient";
  private int           currentGradientID = 0;
  
  private static String txPrefix          = "Texture";
  private int           currentTextureID  = 0;
  

  private String paintToString()
  {
    if( current_paint_type == PAINT_COLOR ) {
      return colToString(currentColour);
    } else if( current_paint_type == PAINT_GRADIENT ) {
      return "url(#" + grPrefix + currentGradientID + ")";
    } else {
      return "url(#" + txPrefix + currentTextureID + ")";
    }      
  }
  
  /**
   * Gets this graphics context's current color.
   * @return this graphics context's current color.
   */
  public Color getColor() 
  { 
    return (currentColour == null) ? Color.black : currentColour;
  }

  /**
   * Store the current colour. This will be the current 'paint' too, because <code>Color</code> also
   * implements the <code>Paint</code> interface.
   *
   * @param c new colour object
   */
  public void setColor(Color c)
  {
    currentColour = c;
    current_paint_type = PAINT_COLOR;
  }

  /**
   * Sets the <code>Paint</code> attribute for the 
   * <code>Graphics2D</code> context.
   * <p><strong>The <code>TexturePaint</code> requires that the image would
   * should have the "url" property!</strong>
   * @param paint new paint object
   */
  public void setPaint( Paint paint )
  {
    if( paint instanceof Color ) {
      setColor((Color)paint);
    } else if( paint instanceof GradientPaint ) {
      GradientPaint g = (GradientPaint)paint;
      currentGradientID++;
      StringBuffer lin = new StringBuffer();
      lin.append("<linearGradient id=\"" + grPrefix + currentGradientID + "\" ");
      lin.append(pairToString(g.getPoint1(),g.getPoint2()) + ">");
      printlnDef(lin.toString());
      printlnDef("  <stop offset=\"0%\" style=\"stop-color:" + colToString(g.getColor1()) + "\"/>");
      printlnDef("  <stop offset=\"100%\" style=\"stop-color:" + colToString(g.getColor2()) + "\"/>");
      printlnDef("</linearGradient>");
      current_paint_type = PAINT_GRADIENT;
    } else if( paint instanceof TexturePaint && theComponent != null ) {
      TexturePaint tx = (TexturePaint)paint;
      Image img = tx.getImage();
      ImageObserver dummy = new ImageObserver() {
        public boolean imageUpdate(Image img, int infoflags, int x, int y,int width,int height)
        {
          return true;
        }
      };
      Object oo = img.getProperty("url",dummy);

      if( oo != null ) {
        if( oo ==  Image.UndefinedProperty )
        { 
          return;
        } else { 
          try {
            String url = (String)oo;
            Rectangle2D rect = tx.getAnchorRect();
            StringBuffer txt = new StringBuffer();
            txt.append("<pattern id=\"" + txPrefix + currentTextureID + "\" ");
            txt.append(rectToStringTrans((float)rect.getX(),(float)rect.getY(),
                                         (float)rect.getWidth(),(float)rect.getHeight()));
            printlnDef(txt.toString());
            printlnDef("  xlink:href=\"" + url + "\"/>");
            current_paint_type = PAINT_TEXTURE;
          } catch( Exception ee ) { return; }
        }
      } else {
        return;
      }
    }
  }

  /**
   * Returns the current <code>Paint</code> of the 
   * <code>Graphics2D</code> context.
   * @return the current <code>Graphics2D</code> <code>Paint</code>,
   * which defines a color or pattern.
   */
  public Paint getPaint()
  {
    switch(current_paint_type) {
      case PAINT_COLOR:    return currentColour;
      case PAINT_GRADIENT: return currentGradient;
      default            : return null;
    }
  }

  /**
   * Sets the background colour for the <code>Graphics2D</code> context. 
   * @param colour the background colour that is used in
   * subsequent calls to <code>clearRect</code>
   */
  public void setBackground(Color colour)
  {
    backgroundColour = colour;
  }

  /**
   * Returns the background colour used for clearing a region.
   * @return the current <code>Color</code>,
   * which defines the background color.
   * @see #setBackground
   */
  public Color getBackground()
  {
    return backgroundColour;
  }

  /**
   * Only plain paint mode is implemented for SVG. 
   */
  public void setPaintMode() {}

  /**
   * Setting XOR mode.
   * <p><strong>XOR mode is not implemented!</strong>
   *
   */
  public void setXORMode(Color c1)
  {
    System.err.println("Sorry, XOR mode is not implemented!");
  }

  //========================================================================
  
  /**
   * Sets the <code>Stroke</code> for the <code>Graphics2D</code> context.
   * A new group will be created in SVG.
   * <p>The only implementing class of this interface in Java 1.2 is <code>BasicStroke</code>,
   * and this assumption is used in the code (to retrieve linewidth, dash array, etc.). If users
   * add their own <code>Stroke</code> implementations, or if new ones appear in later releases
   * of Java, this has to be extended.
   * @param s the <code>Stroke</code> object to be used to stroke a 
   * <code>Shape</code> during the rendering process
   */
  public void setStroke(Stroke s)
  {
    if( s instanceof BasicStroke ) {
      currentStroke = (BasicStroke) s;
      newGroup();
    }
  }
    
  /**
   * Returns the current <code>Stroke</code>.
   * @see #setStroke
   */
  public Stroke getStroke()
  {
    return currentStroke;
  }
  
  //====================================================================
  
  /**
   * Sets the current composite instance.
   * <p>The only implementing class of this interface in Java 1.2 is <code>AlphaComposite</code>,
   * and this assumption is used in the code (to retrieve alpha values). If users
   * add their own <code>Composite</code> implementations, or if new ones appear in later releases
   * of Java, this has to be extended.
   * <p>The only used attribute of the composite class is the alpha value (through the
   * <code>getAlpha()</code>) which is used to set the opacity. I would not know, at this moment,
   * how to translate the complicated composition rules of Java into SVG statements.
   * @param comp the <code>Composite</code> object to be used for rendering.
   */
  public void setComposite(Composite comp)
  { 
    if( comp instanceof AlphaComposite )
      currentComposite = (AlphaComposite)comp; 
  }

  /**
   * Returns the current <code>Composite</code> in the
   * <code>Graphics2D</code> context.
   * @return the current <code>Graphics2D</code> <code>Composite</code>,
   *              which defines a compositing style.
   * @see #setComposite
   */
  public Composite getComposite() { return currentComposite; }

 
  //====================================================================
  private static String[][] logicals = new String[][] {
    new String[] { "dialoginput", "monospace" },
    new String[] { "dialog",      "sans-serif" },
    new String[] { "serif",       "serif" },
    new String[] { "sansserif",   "sans-serif" },
    new String[] { "monospaced",  "monospace" }
  };
  
  private String fontToString()
  {
    StringBuffer buff = new StringBuffer();
    if( currentFont != null ) {
      String javaname = currentFont.getFamily();
      String svgname  = null;
      // We have to see if this is a 'logical' font name in Java terms
      // and choose the SVG/CSS equivalent
      for( int i = 0; i < logicals.length; i++ ) {
        String[] pair = logicals[i];
        if( javaname.startsWith(pair[0]) == true ) {
          // Bingo
          svgname = pair[1];
          break;
        }
      }
      if( svgname == null ) {
        svgname = "'" + javaname + "'";
      }
      buff.append("font-family:" + svgname + "; ");
      buff.append("font-size:" + currentFont.getSize2D() + "; ");
      if( currentFont.isItalic() ) {
        buff.append("font-style:italic; ");
      }
      if( currentFont.isBold() ) {
        buff.append("font-weight:bold; ");
      }
    }
    buff.append("color:" + paintToString() + ";");
    buff.append(" fill:" + paintToString() + ";");
    return buff.toString();
  }     

  /**
   * Gets the current font.
   * @return  current font.
   */
  public Font getFont()
  {
    return currentFont;
  }

  /**
  * Sets the current font.
  * <p>
  * There are some small problems in implementing this method. Indeed, a java user can also set a "logical"
  * font, ie, using a "logical" name for the font instead of the exact family. Of course, the set of
  * available logical names in Java (Dialog, DialogInput, Monospaces, Serif, SansSerif, and Symbol) is 
  * <em>not</em> identical with SVG (more exactly, CSS). It is not clear in Java what Dialog and
  * DialogInput mean and there is no such symbolic font name in CSS either. On a Windows NT environment
  * Dialog seems to be mapped onto
  * sans serif, and DialogInput seems to be mapped to monospaced, so this is what is used.
  * <p>
  * If "real" font family name is used, it is used without any change for SVG.
  * @param font the new font
  */
  public void setFont(Font font)
  {
    currentFont = font;
  }

  /**
   * Gets the font metrics for the specified font.
   * <p>
   * <strong>This is only partially implemented</strong>. There is no real way of inquiring about
   * font data in SVG, so this method is only approximative. As a help for the user, if the
   * "parent" component has been set, then the font metrics generated by that component is returned.
   * @return    the font metrics for the specified font.
   * @param     f the specified font
   */
  public FontMetrics getFontMetrics(Font f)
  {
    if( theComponent == null )
      return null;
    else {
      try {
        return ((Graphics2D)theComponent.getGraphics()).getFontMetrics(f);
      } catch( Exception e ) {
        return null;
      }
    }
  }

  /** 
   * Renders the text of the specified <code>String</code>, using the 
   * current <code>Font</code> and <code>Paint</code> attributes in the 
   * <code>Graphics2D</code> context. 
   * @param s the string to be dislplayed
   * @param x x coordinate for the start of the string
   * @param y y coordinate for the start of the string
   */
  public void drawString(String s, int x, int y)
  {
    drawString(s,(float)x,(float)y);
  }

  /**
   * Renders the text specified by the specified <code>String</code>, 
   * using the current <code>Font</code> and <code>Paint</code> attributes 
   * in the <code>Graphics2D</code> context. 
   * @param str the string to be dislplayed
   * @param x x coordinate for the start of the string
   * @param y y coordinate for the start of the string
   */
  public void drawString(String s, float x, float y)
  {
    println("<text " + pointToString(x,y) + "style=\"" + fontToString() + "\">" + s + "</text>");
  }
  
  /**
   * Renders the text of the specified iterator, using the
   * <code>Graphics2D</code> context's current <code>Paint</code>. 
   * <p>
   * <strong>This method is not yet implemented.</strong> 
   * @param iterator character iterator
   * @param x x coordinate for the start of the string
   * @param y y coordinate for the start of the string
   */
  public void drawString(AttributedCharacterIterator iterator,int x, int y)
  {
    drawString(iterator,(float)x,(float)y);
  }
                                      

  /**
   * Renders the text of the specified iterator, using the
   * <code>Graphics2D</code> context's current <code>Paint</code>. 
   * <p>
   * <strong>This method is not yet implemented.</strong> 
   * @param iterator character iterator
   * @param x x coordinate for the start of the string
   * @param y y coordinate for the start of the string
   */
  public void drawString(AttributedCharacterIterator iterator,float x, float y)
  {
    ;
  }

  /**
   * Renders the text of the specified 
   * GlyphVector using
   * the <code>Graphics2D</code> context's rendering attributes.
   * <p>
   * <strong>This method is not yet implemented.</strong> 
   * @param g glyph vector
   * @param x x coordinate for the start of the string
   * @param y y coordinate for the start of the string
   */
  public void drawGlyphVector(GlyphVector g, float x, float y)
  {
    ;
  }    

  //====================================================================
  
  private Shape currentClipShape = null;

  /**
   * Returns the bounding rectangle of the current clipping area.
   * This method refers to the user clip, which is independent of the
   * clipping associated with device bounds and window visibility.  
   * If no clip has previously been set, or if the clip has been 
   * cleared using <code>setClip(null)</code>, this method returns
   * <code>null</code>.
   * @return      the bounding rectangle of the current clipping area,
   *              or <code>null</code> if no clip is set.
   */
  public Rectangle getClipBounds()
  {
    if( currentClipShape == null ) return null;
    else                           return currentClipShape.getBounds();
  }

  /** 
   * Intersects the current clip with the specified rectangle.
   * The resulting clipping area is the intersection of the current
   * clipping area and the specified rectangle.  If there is no 
   * current clipping area, either because the clip has never been 
   * set, or the clip has been cleared using <code>setClip(null)</code>, 
   * the specified rectangle becomes the new clip.
   * <p>
   * <strong>The intersection of clips with existing shapes is not implemented yet</strong>.
   * In other words, this method is void, unless the current clip shape is not yet
   * set, in which case it is equivalent to a <code>setClip</code> call.
   * @param x upper left hand corner x coordinate
   * @param y upper left hand corner y coordinate
   * @param width width of the rectangle
   * @param height height of the rectangle
   */
  public void clipRect(int x, int y, int width, int height)
  {
    if( currentClipShape == null ) {
      setClip(x,y,width,height);
    } else {
      ;
    }
  }

  /**
   * Sets the current clip to the rectangle specified by the given
   * coordinates.  This method sets the user clip, which is 
   * independent of the clipping associated with device bounds
   * and window visibility. 
   * In SVG, a <code>clipPath</code> is generated. 
   * @param x upper left hand corner x coordinate
   * @param y upper left hand corner y coordinate
   * @param width width of the rectangle
   * @param height height of the rectangle
   */
  public void setClip(int x, int y, int width, int height)
  {
    currentClipShape = new Rectangle(x,y,width,height);    
    println("<clipPath>");
    println("  <rect " + rectToString((float)x,(float)y,(float)width,(float)height) + "/>");
    println("</clipPath>");
  }

  /**
   * Gets the current clipping area.
   */
  public Shape getClip() { return currentClipShape; }

  /**
   * Sets the current clipping area to an arbitrary clip shape.
   * <p>
   * In SVG, a <code>clipPath</code> is generated.
   * @param clip the clip shape
   */
  public void setClip(Shape clip)
  {
    currentClipShape = clip;
    println("<clipPath>");
    if( clip != null ) {
      println("  <path " + shapeToString(clip,true) + "/>");
    } else {
      println("  <rect " + rectToString(0.0f,0.0f,(float)width,(float)height) + "/>");
    }
    println("</clipPath>");
  }

  //====================================================================
  
  /**
   * Copies an area of the component by a distance specified by 
   * <code>dx</code> and <code>dy</code>.
   * <p>
   * <strong>This method is not yet implemented.</strong> 
   * @param       x the <i>x</i> coordinate of the source rectangle.
   * @param       y the <i>y</i> coordinate of the source rectangle.
   * @param       width the width of the source rectangle.
   * @param       height the height of the source rectangle.
   * @param       dx the horizontal distance to copy the pixels.
   * @param       dy the vertical distance to copy the pixels.
   */
  public void copyArea(int x, int y, int width, int height, int dx, int dy)
  {
    ;
  }

  //==========================================================================

    
  /** 
   * Draws a line. 
   * @param   x1  the first point's <i>x</i> coordinate.
   * @param   y1  the first point's <i>y</i> coordinate.
   * @param   x2  the second point's <i>x</i> coordinate.
   * @param   y2  the second point's <i>y</i> coordinate.
   */
  public void drawLine(int x1, int y1, int x2, int y2)
  {
    println("<line " + strokeStyle());
    println("  " + pairToString((float)x1,(float)y1,(float)x2,(float)y2) + "/>");
  }    

  /**
   * Strokes the outline of a <code>Shape</code> using the settings of the 
   * current <code>Graphics2D</code> context.  
   * @param s the <code>Shape</code> to be rendered
   */
  public void draw(Shape s)
  {
    println("<path " + strokeStyle());
    println("  " + shapeToString(s,false) + "/>");
  }
  
  /**
   * Fills the interior of a <code>Shape</code> using the settings of the   
   * <code>Graphics2D</code> context. .
   * @param s the <code>Shape</code> to be filled
   */
  public void fill(Shape s)
  {
    println("<path " + fillStyle());
    println("  " + shapeToString(s,true) + "/>");
  }

  /** 
   * Draws the specified rectangle. 
   * @param         x   the <i>x</i> coordinate 
   *                         of the rectangle to be filled.
   * @param         y   the <i>y</i> coordinate 
   *                         of the rectangle to be filled.
   * @param         width   the width of the rectangle to be filled.
   * @param         height   the height of the rectangle to be filled.
   */
  public void drawRect(int x, int y, int width, int height)
  {
    println("<rect " + strokeStyle());
    println("  " + rectToString((float)x,(float)y,(float)width,(float)height) + "/>");
  }
  
  /** 
   * Fills the specified rectangle. 
   * @param         x   the <i>x</i> coordinate 
   *                         of the rectangle to be filled.
   * @param         y   the <i>y</i> coordinate 
   *                         of the rectangle to be filled.
   * @param         width   the width of the rectangle to be filled.
   * @param         height   the height of the rectangle to be filled.
   */
  public void fillRect(int x, int y, int width, int height)
  {
    println("<rect " + fillStyle());
    println("  " + rectToString((float)x,(float)y,(float)width,(float)height) + "/>");
  }

  /** 
   * Clears the specified rectangle by filling it with a specific colour. 
   * @param       x the <i>x</i> coordinate of the rectangle to clear.
   * @param       y the <i>y</i> coordinate of the rectangle to clear.
   * @param       width the width of the rectangle to clear.
   * @param       height the height of the rectangle to clear.
   */
  private void clearRect(int x, int y, int width, int height, Color col)
  {
    println("<rect style=\"stroke:none; fill:" + colToString(col) + "\"");
    println("  " + rectToString((float)x,(float)y,(float)width,(float)height) + "/>");
  }
    
  /** 
   * Clears the specified rectangle by filling it with the background
   * color of the current drawing surface. 
   * @param       x the <i>x</i> coordinate of the rectangle to clear.
   * @param       y the <i>y</i> coordinate of the rectangle to clear.
   * @param       width the width of the rectangle to clear.
   * @param       height the height of the rectangle to clear.
   */
  public void clearRect(int x, int y, int width, int height)
  {
    clearRect(x,y,width,height,backgroundColour);
  }

  /** 
   * Draws an outlined round-cornered rectangle using this graphics 
   * context's current color. 
   * <p>
   * <strong>At present, the rounded rectangle is replaced by a plain rectangle.</strong> (SVG does
   * not have the notion of rounded rectangle, and an appropriate path should be calculated instead,
   * which has not been done.)
   * @param      x the <i>x</i> coordinate of the rectangle to be drawn.
   * @param      y the <i>y</i> coordinate of the rectangle to be drawn.
   * @param      width the width of the rectangle to be drawn.
   * @param      height the height of the rectangle to be drawn.
   * @param      arcWidth the horizontal diameter of the arc 
   *                    at the four corners.
   * @param      arcHeight the vertical diameter of the arc 
   *                    at the four corners.
   */
  public void drawRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight)
  {
    drawRect(x,y,width,height);
  }   

  /** 
   * Fills the specified rounded corner rectangle with the current color.
   * <p>
   * <strong>At present, the rounded rectangle is replaced by a plain rectangle.</strong> (SVG does
   * not have the notion of rounded rectangle, and an appropriate path should be calculated instead,
   * which has not been done.)
   * @param       x the <i>x</i> coordinate of the rectangle to be filled.
   * @param       y the <i>y</i> coordinate of the rectangle to be filled.
   * @param       width the width of the rectangle to be filled.
   * @param       height the height of the rectangle to be filled.
   * @param       arcWidth the horizontal diameter 
   *                     of the arc at the four corners.
   * @param       arcHeight the vertical diameter 
   *                     of the arc at the four corners.
   */
  public void fillRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight)
  {
    fillRect(x,y,width,height);
  }

  /** 
   * Draws the outline of an oval.
   * <p> 
   * An <code>ellipse</code> statement is generated in SVG.
   * @param       x the <i>x</i> coordinate of the upper left 
   *                     corner of the oval to be drawn.
   * @param       y the <i>y</i> coordinate of the upper left 
   *                     corner of the oval to be drawn.
   * @param       width the width of the oval to be drawn.
   * @param       height the height of the oval to be drawn.
   */
  public void drawOval(int x, int y, int width, int height)
  {
    println("<ellipse " + strokeStyle());
    println("  " + ovalToString((float)x,(float)y,(float)width,(float)height) + "/>");
  }    

  /** 
   * Fills an oval bounded by the specified rectangle with the
   * current color.
   * <p> 
   * An <code>ellipse</code> statement is generated in SVG.
   * @param       x the <i>x</i> coordinate of the upper left corner 
   *                     of the oval to be filled.
   * @param       y the <i>y</i> coordinate of the upper left corner 
   *                     of the oval to be filled.
   * @param       width the width of the oval to be filled.
   * @param       height the height of the oval to be filled.
   */
  public void fillOval(int x, int y, int width, int height)
  {
    println("<ellipse " + fillStyle());
    println("  " + ovalToString((float)x,(float)y,(float)width,(float)height) + "/>");
  }

  /**
   * Draws the outline of a circular or elliptical arc 
   * covering the specified rectangle.
   * @param        x the <i>x</i> coordinate of the 
   *                    upper-left corner of the arc to be drawn.
   * @param        y the <i>y</i>  coordinate of the 
   *                    upper-left corner of the arc to be drawn.
   * @param        width the width of the arc to be drawn.
   * @param        height the height of the arc to be drawn.
   * @param        startAngle the beginning angle.
   * @param        arcAngle the angular extent of the arc, 
   *                    relative to the start angle.
   */
  public void drawArc(int x, int y, int width, int height,int startAngle, int arcAngle)
  {
    println("<path " + strokeStyle());
    println(arcToString((float)x,(float)y,(float)width,(float)height,
                        (float)startAngle,(float)arcAngle,false) + "/>");
  }

  /** 
   * Fills a circular or elliptical arc covering the specified rectangle.
   * @param        x the <i>x</i> coordinate of the 
   *                    upper-left corner of the arc to be filled.
   * @param        y the <i>y</i>  coordinate of the 
   *                    upper-left corner of the arc to be filled.
   * @param        width the width of the arc to be filled.
   * @param        height the height of the arc to be filled.
   * @param        startAngle the beginning angle.
   * @param        arcAngle the angular extent of the arc, 
   *                    relative to the start angle.
   * @see         java.awt.Graphics#drawArc
   */
  public void fillArc(int x, int y, int width, int height, int startAngle, int arcAngle)
  {
    println("<path " + fillStyle());
    println(arcToString((float)x,(float)y,(float)width,(float)height,
                        (float)startAngle,(float)arcAngle,true) + "/>");
  }

  /** 
   * Draws a sequence of connected lines defined by 
   * arrays of <i>x</i> and <i>y</i> coordinates. 
   * 
   * @param       xPoints an array of <i>x</i> points
   * @param       yPoints an array of <i>y</i> points
   * @param       nPoints the total number of points
   * @see         java.awt.Graphics#drawPolygon(int[], int[], int)
   */
  public void drawPolyline(int[] xPoints, int[] yPoints, int nPoints)
  {
    println("<polyline " + strokeStyle());
    println(pointsToString(xPoints,yPoints,nPoints) + "/>");
  }

  /** 
   * Draws a closed polygon defined by 
   * arrays of <i>x</i> and <i>y</i> coordinates. 
   * Each pair of (<i>x</i>,&nbsp;<i>y</i>) coordinates defines a point.
   * 
   * @param        xPoints   a an array of <code>x</code> coordinates.
   * @param        yPoints   a an array of <code>y</code> coordinates.
   * @param        nPoints   a the total number of points.
   */
  public void drawPolygon(int[] xPoints, int[] yPoints, int nPoints)
  {
    println("<polygon " + strokeStyle());
    println(pointsToString(xPoints,yPoints,nPoints) + "/>");
  }

  /** 
   * Fills a closed polygon defined by 
   * arrays of <i>x</i> and <i>y</i> coordinates. 
   * @param        xPoints   a an array of <code>x</code> coordinates.
   * @param        yPoints   a an array of <code>y</code> coordinates.
   * @param        nPoints   a the total number of points.
   * @see          java.awt.Graphics#drawPolygon(int[], int[], int)
   */
  public void fillPolygon(int[] xPoints, int[] yPoints, int nPoints)
  {
    println("<polygon " + fillStyle());
    println(pointsToString(xPoints,yPoints,nPoints) + "/>");
  }

  //==============================================================================

  private boolean drawImage(Image img, float x, float y, float width, float height,
                            ImageObserver observer)
  {
    Object oo = img.getProperty("url",observer);
    if( oo != null ) {
      if( oo ==  Image.UndefinedProperty )
      { 
        return false;
      } else { 
        try {
          String url = (String)oo;
          println("<image " + rectToString(x,y,width,height));
          println("  xlink:href=\"" + url + "\"/>");
          return true;
        } catch( Exception ee ) { return false; }
      }
    } else {
      return false;
    }
  }
    

  /** 
   * Draws as much of the specified image as is currently available.
   * <p>
   * <strong>The image has to have the property <code>"url"</code>.</strong> See comments
   * on the class level for further details.
   * @param    img the specified image to be drawn.
   * @param    x   the <i>x</i> coordinate.
   * @param    y   the <i>y</i> coordinate.
   * @param    obs    object to be notified as more of 
   *                          the image is converted.
   */
  public boolean drawImage(Image img, int x, int y, ImageObserver obs)
  {
    return drawImage(img,(float)x,(float)y,(float)img.getWidth(obs),(float)img.getHeight(obs),obs);
  }

  /**
   * Draws as much of the specified image as has already been scaled
   * to fit inside the specified rectangle.
   * <p>
   * <strong>The image has to have the property <code>"url"</code>.</strong> See comments
   * on the class level for further details.
   * @param    img    the specified image to be drawn.
   * @param    x      the <i>x</i> coordinate.
   * @param    y      the <i>y</i> coordinate.
   * @param    width  the width of the rectangle.
   * @param    height the height of the rectangle.
   * @param    observer    object to be notified as more of 
   *                          the image is converted.
   */
  public boolean drawImage(Image img, int x, int y, int width, int height, ImageObserver obs)
  {
    return drawImage(img,(float)x,(float)y,(float)width,(float)height,obs);
  }
    
  /** 
   * Draws as much of the specified image as is currently available.
   * <p>
   * <strong>The image has to have the property <code>"url"</code>.</strong> See comments
   * on the class level for further details.
   * @param    img    the specified image to be drawn.
   * @param    x      the <i>x</i> coordinate.
   * @param    y      the <i>y</i> coordinate.
   * @param    bgcolor the background color to paint under the
   *                         non-opaque portions of the image.
   * @param    observer    object to be notified as more of 
   *                          the image is converted.
   */
  public boolean drawImage(Image img, int x, int y, Color bgcolor, ImageObserver obs)
  {
    int width  = img.getWidth(obs);
    int height = img.getHeight(obs);
    return drawImage(img,x,y,width,height,bgcolor,obs);
  }

  /**
   * Draws as much of the specified image as has already been scaled
   * to fit inside the specified rectangle.
   * <p>
   * <strong>The image has to have the property <code>"url"</code>.</strong> See comments
   * on the class level for further details.
   * 
   * @param    img       the specified image to be drawn.
   * @param    x         the <i>x</i> coordinate.
   * @param    y         the <i>y</i> coordinate.
   * @param    width     the width of the rectangle.
   * @param    height    the height of the rectangle.
   * @param    bgcolor   the background color to paint under the
   *                         non-opaque portions of the image.
   * @param    observer    object to be notified as more of 
   *                          the image is converted.
   */
  public boolean drawImage(Image img, int x, int y, int width, int height,Color bgcolor,
                           ImageObserver obs)
  {
    clearRect(x,y,width,height,bgcolor);
    return drawImage(img,x,y,width,height,obs);
  }    
    
  /**
   * Draws as much of the specified area of the specified image as is
   * currently available, scaling it on the fly to fit inside the
   * specified area of the destination drawable surface. Transparent pixels 
   * do not affect whatever pixels are already there.
   * <p>
   * <strong>Cutting part of the image is not yet implemented</strong>. (It would probably
   * require to use some of the image filters defined in SVG, but this has not
   * been checked.)
   * @param       img the specified image to be drawn
   * @param       dx1 the <i>x</i> coordinate of the first corner of the
   *                    destination rectangle.
   * @param       dy1 the <i>y</i> coordinate of the first corner of the
   *                    destination rectangle.
   * @param       dx2 the <i>x</i> coordinate of the second corner of the
   *                    destination rectangle.
   * @param       dy2 the <i>y</i> coordinate of the second corner of the
   *                    destination rectangle.
   * @param       sx1 the <i>x</i> coordinate of the first corner of the
   *                    source rectangle.
   * @param       sy1 the <i>y</i> coordinate of the first corner of the
   *                    source rectangle.
   * @param       sx2 the <i>x</i> coordinate of the second corner of the
   *                    source rectangle.
   * @param       sy2 the <i>y</i> coordinate of the second corner of the
   *                    source rectangle.
   * @param       observer object to be notified as more of the image is
   *                    scaled and converted.
   */
  public boolean drawImage(Image img,
            int dx1, int dy1, int dx2, int dy2,
            int sx1, int sy1, int sx2, int sy2,
            ImageObserver observer)
  {
    return false;
  }

  /**
   * Draws as much of the specified area of the specified image as is
   * currently available, scaling it on the fly to fit inside the
   * specified area of the destination drawable surface. 
   * <p>
   * <strong>Cutting part of the image is not yet implemented</strong>. (It would probably
   * require to use some of the image filters defined in SVG, but this has not
   * been checked.)
   * @param       img the specified image to be drawn
   * @param       dx1 the <i>x</i> coordinate of the first corner of the
   *                    destination rectangle.
   * @param       dy1 the <i>y</i> coordinate of the first corner of the
   *                    destination rectangle.
   * @param       dx2 the <i>x</i> coordinate of the second corner of the
   *                    destination rectangle.
   * @param       dy2 the <i>y</i> coordinate of the second corner of the
   *                    destination rectangle.
   * @param       sx1 the <i>x</i> coordinate of the first corner of the
   *                    source rectangle.
   * @param       sy1 the <i>y</i> coordinate of the first corner of the
   *                    source rectangle.
   * @param       sx2 the <i>x</i> coordinate of the second corner of the
   *                    source rectangle.
   * @param       sy2 the <i>y</i> coordinate of the second corner of the
   *                    source rectangle.
   * @param       bgcolor the background color to paint under the
   *                    non-opaque portions of the image.
   * @param       observer object to be notified as more of the image is
   *                    scaled and converted.
   */
  public boolean drawImage(Image img,
            int dx1, int dy1, int dx2, int dy2,
            int sx1, int sy1, int sx2, int sy2,
            Color bgcolor,
            ImageObserver observer)
  {
    return false;
  }

  /**
   * Renders an image, applying a transform from image space into user space
   * before drawing.
   * <p>
   * The method is implemented by enclosing a usual image call into a
   * group with the right transformation. 
   * <p>
   * <strong>The image has to have the property <code>"url"</code>.</strong> See comments
   * on the class level for further details.
   * @param img the <code>Image</code> to be rendered
   * @param xform the transformation from image space into user space
   * @param obs the ImageObserver
   * to be notified as more of the <code>Image</code>
   * is converted
   * @return <code>true</code> if the image could be displayed, and <code>false</code> otherwise.
   */
  public boolean drawImage(Image img,
                           AffineTransform xform,
                           ImageObserver obs)
  {
    StringBuffer str = new StringBuffer();
    str.append("<g transform=\"matrix(");
    double[] matrix = new double[6];
    xform.getMatrix(matrix);
    for( int i = 0; i < 6; i++ ) {
      str.append(flstr(matrix[i]));
      if( i < 5 ) str.append(",");
    }
    str.append(")\"");
    println(str.toString());
    sp += 2;
    boolean retval = drawImage(img,0,0,obs);
    sp -= 2;
    println("</g>");
    return retval;
  }
    

  /**
   * Renders a <code>BufferedImage</code>.
   * <p>
   * <strong>Handling of Buffered Images is not implemented.</strong>
   * @param op the filter to be applied to the image before rendering
   * @param img the <code>BufferedImage</code> to be rendered
   * @param x,&nbsp;y the location in user space where the upper left
   * corner of the
   * image is rendered
   * @see #transform
   * @see #setTransform
   * @see #setComposite
   * @see #clip
   * @see #setClip
   */
  public void drawImage(BufferedImage img,BufferedImageOp op,int x,int y)
  {
    ;
  }

  /**
   * Draws a rendered image
   * <p><strong>Handling of rendered images is not implemented.</strong>
   * @param img the image to be rendered
   * @param xform the transformation from image space into user space
   * @see #transform
   * @see #setTransform
   * @see #setComposite
   * @see #clip
   * @see #setClip
   */
  public void drawRenderedImage(RenderedImage img,AffineTransform xform)
  {
    ;
  }

  /**
   * Draws a renderable image
   * <p><strong>Handling of renderable images is not implemented.</strong>
   * @param img the image to be rendered
   * @param xform the transformation from image space into user space
   * @see #transform
   * @see #setTransform
   * @see #setComposite
   * @see #clip
   * @see #setClip
   * @see #drawRenderedImage
   */
  public void drawRenderableImage(RenderableImage img, AffineTransform xform)
  {
    ;
  }

  //===========================================================================
  
  /**
   * Checks whether or not the specified <code>Shape</code> intersects 
   * the specified Rectangle, which is in device
   * space. 
   * @return <code>true</code> if there is a hit; <code>false</code> 
   * otherwise.
   */
  public boolean hit(Rectangle rect,Shape s,boolean onStroke)
  {
    return s.intersects(new Rectangle2D.Float((float)rect.x,(float)rect.y,
                                              (float)rect.width,(float)rect.height));
  }



  /**
   * Sets the value of a single preference for the rendering algorithms.
   * <p>
   * <strong>This method has no effect in SVG.</strong>
   * @param hintKey the key of the hint to be set. 
   * @param hintValue the value indicating preferences for the specified
   * hint category.
   * @see RenderingHints
   */
  public void setRenderingHint(RenderingHints.Key hintKey, Object hintValue)
  {
    ;
  }

  /**
   * Returns the value of a single preference for the rendering algorithms.
   * <p>
   * <strong>The returned values are constants and, in a sense, meaningless</strong>.
   * The rendering quality depends on the quality of the SVG browser, and not the 
   * SVG file itself, so the assumption is that the browser is of the highest quality. 
   * @param hintKey the key corresponding to the hint to get. 
   * @return an object representing the value for the specified hint key.
   * Some of the keys and their associated values are defined in the
   * <code>RenderingHints</code> class.
   * @see RenderingHints
   */
  public Object getRenderingHint(RenderingHints.Key hintKey)
  {
    if(hintKey.equals(RenderingHints.KEY_ALPHA_INTERPOLATION)) {
      return RenderingHints.VALUE_ALPHA_INTERPOLATION_DEFAULT; 
    } else if(hintKey.equals(RenderingHints.KEY_ANTIALIASING)) {
      return RenderingHints.VALUE_ANTIALIAS_ON; 
    } else if(hintKey.equals(RenderingHints.KEY_COLOR_RENDERING)) {
      return RenderingHints.VALUE_COLOR_RENDER_QUALITY; 
    } else if(hintKey.equals(RenderingHints.KEY_DITHERING)) {
      return RenderingHints.VALUE_DITHER_ENABLE; 
    } else if(hintKey.equals(RenderingHints.KEY_FRACTIONALMETRICS)) {
      return RenderingHints.VALUE_FRACTIONALMETRICS_ON; 
    } else if(hintKey.equals(RenderingHints.KEY_INTERPOLATION)) {
      return RenderingHints.VALUE_INTERPOLATION_BICUBIC; 
    } else if(hintKey.equals(RenderingHints.KEY_RENDERING)) {
      return RenderingHints.VALUE_RENDER_QUALITY; 
    } else if(hintKey.equals(RenderingHints.KEY_TEXT_ANTIALIASING)) {
      return RenderingHints.VALUE_TEXT_ANTIALIAS_ON;
    } else {
      return null;
    }
  }

  /**
   * Replaces the values of all preferences for the rendering
   * algorithms with the specified <code>hints</code>.
   * <p>
   * <strong>The method is void and, in a sense, meaningless</strong>.
   * The rendering quality depends on the quality of the SVG browser, and not the 
   * SVG file itself, so the assumption is that the browser is of the highest quality. 
   * @param hints the rendering hints to be set
   * @see RenderingHints
   */
  public void setRenderingHints(Map hints)
  {
    ;
  }
  
  /**
   * Sets the values of an arbitrary number of preferences for the
   * rendering algorithms.
   * <p>
   * <strong>The method is void and, in a sense, meaningless</strong>.
   * The rendering quality depends on the quality of the SVG browser, and not the 
   * SVG file itself, so the assumption is that the browser is of the highest quality. 
   * @param hints the rendering hints to be set
   * @see RenderingHints
   */
  public void addRenderingHints(Map hints)
  {
    ;
  }

  /**
   * Gets the preferences for the rendering algorithms.
   * <p>
   * <strong>The returned values are constants and, in a sense, meaningless</strong>.
   * The rendering quality depends on the quality of the SVG browser, and not the 
   * SVG file itself, so the assumption is that the browser is of the highest quality. 
   * @return a reference to an instance of <code>RenderingHints</code>
   * that contains the current preferences.
   * @see RenderingHints
   */
  public RenderingHints getRenderingHints()
  {
    RenderingHints hints = new RenderingHints(
                                 RenderingHints.KEY_ALPHA_INTERPOLATION,
                                 getRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION)
                               );
    hints.put(RenderingHints.KEY_ANTIALIASING,
              getRenderingHint(RenderingHints.KEY_ANTIALIASING)
             );
    hints.put(RenderingHints.KEY_COLOR_RENDERING,
              getRenderingHint(RenderingHints.KEY_COLOR_RENDERING)
             );
    hints.put(RenderingHints.KEY_DITHERING,
              getRenderingHint(RenderingHints.KEY_DITHERING)
             );
    hints.put(RenderingHints.KEY_FRACTIONALMETRICS,
              getRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS)
             );
    hints.put(RenderingHints.KEY_INTERPOLATION,
              getRenderingHint(RenderingHints.KEY_INTERPOLATION)
             );
    hints.put(RenderingHints.KEY_RENDERING,
              getRenderingHint(RenderingHints.KEY_RENDERING)
             );
    hints.put(RenderingHints.KEY_TEXT_ANTIALIASING,
              getRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING)
             );
             
    return hints;
  }

  /**
  * Intersects the current <code>Clip</code> with the interior of the
  * specified <code>Shape</code> and sets the <code>Clip</code> to the
  * resulting intersection. 
  * <p>
  * <strong>This method has not been implemented yet.</strong> 
  * @param s the <code>Shape</code> to be intersected with the current
  *          <code>Clip</code>.  If <code>s</code> is <code>null</code>,
  *          this method clears the current <code>Clip</code>.
  */
  public void clip(Shape s)
  {
    ;
  }

  /**
  * Get the rendering context of the <code>Font</code> within this 
  * <code>Graphics2D</code> context.
  * @return a reference to an instance of FontRenderContext.
  */
  public FontRenderContext getFontRenderContext()
  {
    return new FontRenderContext(new AffineTransform(), true, true);
  }
  
  //===================================================================
  /**
   * Returns the device configuration associated with this 
   * <code>Graphics2D</code>.
   * <p>
   * <strong>The implementation is fake.</strong> Just to give an indication to the
   * user, it returns the graphics configuration of the "parent" (i.e., "real") component,
   * although this might be meaningless for the display of the SVG file.
   * @return the graphics configuration of the "parent" component, if set, and null
   * otherwise.
   */
  public GraphicsConfiguration getDeviceConfiguration()
  {
    if( theComponent == null )
      return null;
    else {
      try {
        return ((Graphics2D)theComponent.getGraphics()).getDeviceConfiguration();
      } catch( Exception e ) {
        return null;
      }
    }
  }
}